#include "miniaudioengine.h"
#define LPF_BIAS            0.9f
#define LPF_CUTOFF_FACTOR   80
#define HPF_CUTOFF_FACTOR   1
#define LPF_ORDER           8

MiniAudioEngine::MiniAudioEngine()
{
    for (int i =0; i < MAX_LAYERS; i++) {
        m_mediaLoaded[i] = false;
        m_currentLayerValues[i].status = Status::Iddle;
        m_currentLayerValues[i].pan = 128;
        m_currentLayerValues[i].pitch = 128;
        m_currentLayerValues[i].vol = 0;
        m_currentLayerValues[i].cursor = 0;
    }
}

void MiniAudioEngine::audioDataCallback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    (void)pInput;
    (void)pDevice;
    ma_result result;
    result = ma_engine_read_pcm_frames((ma_engine*)pDevice->pUserData, pOutput, frameCount, NULL);
    if (result != MA_SUCCESS) {
        cout << "1";
    }
}

void MiniAudioEngine::stopEngine()
{
    for (uint i = 0; i < m_devicesSelected; i++) {
        ma_engine_uninit(&m_engine[i]);
        ma_device_uninit(&m_device[i]);
    }
    ma_context_uninit(&m_context);
    ma_resource_manager_uninit(&m_resourceManager);
}

bool MiniAudioEngine::startEngine()
{
    ma_result result;

    result = this->startContext();
    if (result != MA_SUCCESS) return false;
    result = this->getAllAudioDevices();
    if (result != MA_SUCCESS) return false;
    return true;
}

ma_result MiniAudioEngine::setNodeGraph(int id) {
    ma_result result;

    ma_lpf_node_config lpfNodeConfig = ma_lpf_node_config_init(CHANNELS, SAMPLE_RATE, SAMPLE_RATE / LPF_CUTOFF_FACTOR, LPF_ORDER);
    ma_node_graph *ng = ma_engine_get_node_graph(&m_engine[id]);
    result = ma_lpf_node_init(ng, &lpfNodeConfig, NULL, &m_filterBank[id].lpfNode);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to initialize low pass filter node." << endl;
        return result;
    }
    ma_node *endpoint = ma_engine_get_endpoint(&m_engine[id]);
    // ToDo: ampliar dimensión a m_filterBank con las capas
    ma_hpf_node_config hpfNodeConfig = ma_hpf_node_config_init(CHANNELS, SAMPLE_RATE, SAMPLE_RATE / HPF_CUTOFF_FACTOR, LPF_ORDER);
    result = ma_hpf_node_init(ng, &hpfNodeConfig, NULL, &m_filterBank[id].hpfNode);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to initialize high pass filter node." << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&m_filterBank[id].lpfNode, 0, &m_filterBank[id].hpfNode, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach low pass filter node." << endl;
        return result;
    }
    // ToDo: add peak filters
    result = ma_node_attach_output_bus(&m_filterBank[id].hpfNode, 0, endpoint, 0);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to attach low pass filter node." << endl;
        return result;
    }
/*
    result = ma_node_set_state(&m_filterBank[id].lpfNode, ma_node_state::ma_node_state_started);
    if (result != MA_SUCCESS) {
        cout << "ERROR " << result << ": Failed to set state to filter node." << endl;
        return result;
    }*/
    return (result);
}

bool MiniAudioEngine::startDevice(uint *systemId, uint nb)
{
    ma_result result = MA_SUCCESS;
    ma_device_config deviceConfig;
    ma_engine_config engineConfig;

    m_devicesSelected = nb;
    for (uint internalId = 0; internalId < nb; internalId++) {
        deviceConfig = ma_device_config_init(ma_device_type_playback);
        deviceConfig.playback.pDeviceID = &m_pPlaybackDeviceInfos[systemId[internalId]].id;
        deviceConfig.playback.format    = m_resourceManager.config.decodedFormat;
        deviceConfig.playback.channels  = 0;
        deviceConfig.sampleRate         = m_resourceManager.config.decodedSampleRate;
        deviceConfig.dataCallback       = audioDataCallback;
        deviceConfig.pUserData          = &m_engine[internalId];
        result = ma_device_init(&m_context, &deviceConfig, &m_device[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to initialize audio device " << m_pPlaybackDeviceInfos[*systemId].name << endl;
            return false;
        }
        engineConfig                  = ma_engine_config_init();
        engineConfig.pDevice          = &m_device[internalId];
        engineConfig.pResourceManager = &m_resourceManager;
        engineConfig.noAutoStart      = MA_TRUE;
        result = ma_engine_init(&engineConfig, &m_engine[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to initialize audio engine" << endl;
            return false;
        }
        result = this->setNodeGraph(internalId);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to set node graph " << systemId[internalId] << endl;
            return false;
        }
        result = ma_engine_start(&m_engine[internalId]);
        if (result != MA_SUCCESS) {
            cout << "Error " << result << ": Failed to start audio engine" << systemId[internalId] << endl;
            return false;
        }
        cout << "Initialized Audio Device. internalId: " << internalId << " systemId: " << systemId[internalId] << " " << m_pPlaybackDeviceInfos[systemId[internalId]].name << endl;
    }
    return true;
}

ma_result MiniAudioEngine::startContext()
{
    ma_result result;

    m_resourceManagerConfig = ma_resource_manager_config_init();
    m_resourceManagerConfig.decodedFormat     = FORMAT;
    m_resourceManagerConfig.decodedChannels   = CHANNELS;
    m_resourceManagerConfig.decodedSampleRate = SAMPLE_RATE;
    m_resourceManagerConfig.jobThreadCount    = 4;
    result = ma_resource_manager_init(&m_resourceManagerConfig, &m_resourceManager);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to initialize audio resource manager." << endl;
        return result;
    }
    result = ma_context_init(NULL, 0, NULL, &m_context);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to initialize audio context." << endl;
    }
    return result;
}

// enum all audio devices in system
ma_result MiniAudioEngine::getAllAudioDevices()
{
    ma_result result;

    result = ma_context_get_devices(&m_context, &m_pPlaybackDeviceInfos, &m_playbackDeviceCount, NULL, NULL);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to enumerate playback devices." << endl;
        ma_context_uninit(&m_context);
        return result;
    }
    cout << "Audio devices available:" << endl;
    for (ma_uint32 iAvailableDevice = 0; iAvailableDevice < m_playbackDeviceCount; iAvailableDevice += 1) {
        cout << iAvailableDevice << " : " << m_pPlaybackDeviceInfos[iAvailableDevice].name << endl;
    }
    return result;
}

ma_result MiniAudioEngine::loadMedia(int layer, char *file, uint audioDevice)
{
    ma_result result;

    // ToDo: ver si s puede attach dos dispositivos a la vez. si no:
    // enchufar a un splitter al sonido y attach cada uno de los lados.
    // iniciar un sonido por cada capa
    if (m_mediaLoaded[layer] == true)
    {
        ma_sound_uninit(&m_currentSound[layer]);
        m_mediaLoaded[layer] = false;
    }
    result = ma_sound_init_from_file(&m_engine[audioDevice], file, \
                MA_SOUND_FLAG_NO_SPATIALIZATION \
                , NULL, NULL, &m_currentSound[layer]);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to load file " << file << endl;
        return result;
    }
    result = ma_node_attach_output_bus(&m_currentSound[layer], 0, &m_filterBank[audioDevice].lpfNode, 0);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to attach output bus " << audioDevice << endl;
        //return result;
    }
    // ToDo: ampliar dimensión a m_filterBank con las capas
    // attach las capas en los dispositivos
    // master cada capa? al principio o final del filtro?
    m_mediaLoaded[layer] = true;
    this->refreshValues(layer);
    m_currentLayerValues[layer].media = file;
    return result;
}

float MiniAudioEngine::getDuration(int layer)
{
    ma_result result;
    float ret;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS) {
        return result;
    }
    return (ret * 1000);
}

float MiniAudioEngine::getCursor(int layer)
{
    ma_result result;
    float ret = 0;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_cursor_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS)
    {
        cout << "Error" << result << ": Can not get cursor " << layer << endl;
        ret = MA_ERROR;
    }
    return ret;
}

ma_result MiniAudioEngine::printFormatInfo(int layer)
{
    ma_format format;
    ma_uint32 channels;
    ma_uint32 sampleRate;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    ma_result result = ma_sound_get_data_format(&m_currentSound[layer], \
                                    &format, &channels, &sampleRate, NULL, 0);
    if (result != MA_SUCCESS) {
        cout << "Error " << result << ": Failed to get data format " << layer;
        cout << endl;
    } else {
        cout << "Layer:" << layer << " ";
        cout << m_currentLayerValues[layer].media.toLatin1().data();
        cout << " samples/sec:" << sampleRate << " format:" <<  format;
        cout << " channels:" << channels << endl;
    }
    return result;
}

// Expects between 0 and 1 vol value
void MiniAudioEngine::volChanged(int layer, float vol)
{
    if (m_mediaLoaded[layer] == false)
        return;
    if (vol >= 1)
        vol = 0.99f;
    ma_sound_group_set_fade_in_milliseconds(&m_currentSound[layer], -1, vol, FADE_TIME);
    m_currentLayerValues[layer].vol = vol;
}

void MiniAudioEngine::panChanged(int layer, float value)
{
    float result;

    if (m_mediaLoaded[layer] == false)
        return;
    result = (value / 128.0) - 1.0;
    ma_sound_group_set_pan(&m_currentSound[layer], result);
    m_currentLayerValues[layer].pan = value;
}

void MiniAudioEngine::pitchChanged(int layer, float value)
{
    float pitch;

    if (m_mediaLoaded[layer] == false)
        return;
    pitch = value / 128.0;
    ma_sound_group_set_pitch(&m_currentSound[layer], pitch);
    m_currentLayerValues[layer].pitch = value;
}

ma_result MiniAudioEngine::playbackChanged(int layer, Status status)
{
    ma_result result = MA_SUCCESS;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    bool loop = false;
    switch (status) {
        case Status::Paused:
            result = ma_sound_stop_with_fade_in_milliseconds(&m_currentSound[layer], FADE_TIME);
            break;
        case Status::Stopped:
            ma_sound_stop_with_fade_in_milliseconds(&m_currentSound[layer], FADE_TIME);
            result = this->seekToCursor(layer, m_currentLayerValues[layer].cursor);
            break;
        case Status::PlayingLoop:
            loop = true;
        case Status::PlayingOnce:
        case Status::PlayingFolder:
        case Status::PlayingFolderLoop:
        case Status::PlayingFolderRandom:
            ma_sound_set_stop_time_in_milliseconds(&m_currentSound[layer], ~(ma_uint64)0);
            ma_sound_set_looping(&m_currentSound[layer], loop);
            result = ma_sound_start(&m_currentSound[layer]);
            //this->volChanged(layer, m_currentLayerValues[layer].vol); // glitch when seek to cursor, how flush the audio buffer?
        default:
            break;
    }
    if (result == MA_SUCCESS)
        m_currentLayerValues[layer].status = status;
    return result;
}

ma_result MiniAudioEngine::seekToCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;
    ma_uint64 end, start;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_pcm_frames(&m_currentSound[layer], &end);
    if (result != MA_SUCCESS) { return result; }
    start = (cursor * end) / 65025;
    result = ma_sound_seek_to_pcm_frame(&m_currentSound[layer], start);
    //result = ma_data_source_set_loop_point_in_pcm_frames(&m_currentSound[layer], start, end); // this do nothing here, it must be done after set_looping or start?
    return (result);
}

ma_result MiniAudioEngine::setCursor(int layer, int cursor)
{
    ma_result result = MA_SUCCESS;

    m_currentLayerValues[layer].cursor = cursor;
    result = this->seekToCursor(layer, cursor);
    return (result);
}

Status MiniAudioEngine::getStatus(int layer)
{
    return m_currentLayerValues[layer].status;
}

void MiniAudioEngine::refreshValues(int layer)
{
    this->seekToCursor(layer, m_currentLayerValues[layer].cursor);
    this->panChanged(layer, m_currentLayerValues[layer].pan);
    this->volChanged(layer, m_currentLayerValues[layer].vol);
    this->pitchChanged(layer, m_currentLayerValues[layer].pitch);
    this->playbackChanged(layer, m_currentLayerValues[layer].status);
}
